import typing

from annadb.data_types.journal import QuerySet
from annadb.query.base import FindInterface
from annadb.query.types import Offset

if typing.TYPE_CHECKING:
    from annadb.collection import Collection


class OffsetQuery(FindInterface):
    def __init__(
        self, number: int, query_set: QuerySet, collection: "Collection"
    ):
        self.collection = collection
        self.query_set = query_set
        self.query_set.push(Offset(number))



from annadb.constants import (
    STRING,
    NUMBER,
    BOOL,
    NULL,
    COLLECTION_NAME,
    PATH_TO_VALUE,
    DELETED,
    STORAGE_MAP,
    STORAGE_VECTOR,
    INSERT_QUERY,
    FIND_QUERY,
    GET_QUERY,
    UPDATE_QUERY,
    EQ_OPERATOR,
    NEQ_OPERATOR,
    GT_OPERATOR,
    GTE_OPERATOR,
    LT_OPERATOR,
    LTE_OPERATOR,
    AND_OPERATOR,
    OR_OPERATOR,
    NOT_OPERATOR,
    SET_OPERATOR,
    INC_OPERATOR,
    RESPONSE_OBJECTS,
    RESPONSE_IDS,
    ROOT,
    DELETE_QUERY,
    RESPONSE_PRIMITIVE,
    OK_RESPONSE,
    ERROR_RESPONSE,
    INSERT_META,
    GET_META,
    QUERY_RESPONSE,
    UTS_PREFIX,
    FIND_META,
    UPDATE_META,
    DELETE_META,
    SORT_QUERY,
    ASC_OPERATOR,
    DESC_OPERATOR,
    QUERY_SET,
    LIMIT_QUERY,
    OFFSET_QUERY,
)
from annadb.data_types.map import Map
from annadb.data_types.primitive import (
    String,
    Number,
    Bool,
    Null,
    CollectionName,
    Deleted,
    Link,
    UTS,
)
from annadb.data_types.vector import Vector
from annadb.query.find.operators import Eq, Neq, Gt, Gte, Lt, Lte, And, Or, Not
from annadb.query.path import Path
from annadb.query.query_set import QuerySet
from annadb.query.sort.operators import Asc, Desc
from annadb.query.types import (
    Insert,
    Find,
    Get,
    Update,
    Delete,
    Sort,
    Limit,
    Offset,
)
from annadb.query.update.operators import Set, Inc
from annadb.response import (
    Objects,
    IDs,
    ResponsePrimitive,
    OkResponse,
    ErrorResponse,
    GetMeta,
    InsertMeta,
    FindMeta,
    UpdateMeta,
    DeleteMeta,
    QueryResponse,
)

types_map = {
    STRING: String,
    NUMBER: Number,
    UTS_PREFIX: UTS,
    BOOL: Bool,
    NULL: Null,
    COLLECTION_NAME: CollectionName,
    PATH_TO_VALUE: Path,
    DELETED: Deleted,
    STORAGE_MAP: Map,
    STORAGE_VECTOR: Vector,
    # QUERIES
    QUERY_SET: QuerySet,
    INSERT_QUERY: Insert,
    FIND_QUERY: Find,
    GET_QUERY: Get,
    UPDATE_QUERY: Update,
    DELETE_QUERY: Delete,
    SORT_QUERY: Sort,
    LIMIT_QUERY: Limit,
    OFFSET_QUERY: Offset,
    # FIND OPERATORS
    EQ_OPERATOR: Eq,
    NEQ_OPERATOR: Neq,
    GT_OPERATOR: Gt,
    GTE_OPERATOR: Gte,
    LT_OPERATOR: Lt,
    LTE_OPERATOR: Lte,
    AND_OPERATOR: And,
    OR_OPERATOR: Or,
    NOT_OPERATOR: Not,
    # UPDATE OPERATORS
    SET_OPERATOR: Set,
    INC_OPERATOR: Inc,
    # SORT OPERATORS
    ASC_OPERATOR: Asc,
    DESC_OPERATOR: Desc,
    # RESPONSE
    RESPONSE_OBJECTS: Objects,
    RESPONSE_IDS: IDs,
    RESPONSE_PRIMITIVE: ResponsePrimitive,
    OK_RESPONSE: OkResponse,
    ERROR_RESPONSE: ErrorResponse,
    GET_META: GetMeta,
    INSERT_META: InsertMeta,
    FIND_META: FindMeta,
    UPDATE_META: UpdateMeta,
    DELETE_META: DeleteMeta,
    QUERY_RESPONSE: QueryResponse,
    # OTHER
    ROOT: Path,
}


def build_primitive(item):
    cls = types_map.get(item.prefix, Link)
    if issubclass(cls, Link):
        return cls(item.value, prefix=item.prefix)
    elif issubclass(cls, String):
        return cls.from_serialized(item.value)
    else:
        return cls(item.value)


def factory(item):
    if i := item.primitive():
        return build_primitive(i)
    if i := item.modifier():
        cls = types_map[i.prefix]
        return cls(factory(i.get_value()))
    if i := item.map():
        obj = types_map[i.prefix]()
        for k, v in i.value:
            obj[build_primitive(k)] = factory(v)
        return obj
    if i := item.vector():
        cls = types_map[i.prefix]
        return cls(*[factory(j) for j in i.value])



from annadb.constants import ASC_OPERATOR, DESC_OPERATOR
from annadb.data_types.modifier import ModifierBase


class Asc(ModifierBase):
    prefix = ASC_OPERATOR


class Desc(ModifierBase):
    prefix = DESC_OPERATOR



from annadb.response import ErrorResponse


class BaseQuery:
    def run(self):
        res = self.collection.conn.send_query(self.query_set)
        if not isinstance(res, ErrorResponse):
            res = res[0]
        return res


class FindInterface(BaseQuery):
    def get(self, *args):
        from annadb.query.get.query import GetQuery

        return GetQuery(
            *args, query_set=self.query_set, collection=self.collection
        )

    def find(self, *args):
        from annadb.query.find.query import FindQuery

        return FindQuery(
            *args, query_set=self.query_set, collection=self.collection
        )

    def delete(self):
        from annadb.query.delete.query import DeleteQuery

        return DeleteQuery(
            query_set=self.query_set, collection=self.collection
        )

    def sort(self, *args):
        from annadb.query.sort.query import SortQuery

        return SortQuery(
            *args, query_set=self.query_set, collection=self.collection
        )

    def limit(self, number: int):
        from annadb.query.limit.query import LimitQuery

        return LimitQuery(
            number, query_set=self.query_set, collection=self.collection
        )

    def offset(self, number: int):
        from annadb.query.offset.query import OffsetQuery

        return OffsetQuery(
            number, query_set=self.query_set, collection=self.collection
        )

    def update(self, *args):
        from annadb.query.update.query import UpdateQuery

        return UpdateQuery(
            *args, query_set=self.query_set, collection=self.collection
        )



from urllib.parse import urlparse

import zmq

from annadb.collection import Collection
from annadb.dump import to_str
from annadb.data_types.journal import Journal
from annadb.transaction import Transaction

context = zmq.Context()


class Connection:
    def __init__(self, user_name: str, password: str, host: str, port: int):
        self.user_name = user_name
        self.password = password
        self.host = host
        self.port = port

        self.socket = context.socket(zmq.REQ)
        self.socket.connect(f"tcp://{host}:{port}")

    def send_query(self, *args, value_only: bool = True) -> Journal:
        data = []
        for query_set in args:
            if isinstance(query_set, str):
                if query_set.endswith(";"):
                    query_set = query_set[:-1]
                data.append(query_set)
            else:
                data.append(to_str(query_set))
        data = ";".join(data)
        self.socket.send_string(data)
        raw_response = self.socket.recv_string()
        response = Journal.deserialize(raw_response)
        if value_only:
            return response[0].value
        return response

    @classmethod
    def from_connection_string(cls, uri: str) -> "Connection":
        res = urlparse(uri)
        return cls(
            user_name=res.username,
            password=res.password,
            host=res.hostname,
            port=res.port,
        )

    def close(self):
        self.socket.close()

    def __getitem__(self, item) -> Collection:
        return Collection(name=item, conn=self)

    def new_transaction(self) -> Transaction:
        return Transaction(self)



from annadb.constants import (
    RESPONSE_IDS,
    OK_RESPONSE,
    GET_META,
    INSERT_META,
    FIND_META,
    UPDATE_META,
    DELETE_META,
    QUERY_RESPONSE,
    RESPONSE_OBJECTS,
    RESPONSE_PRIMITIVE,
    ERROR_RESPONSE,
)
from annadb.data_types.map import MapUnique
from annadb.data_types.primitive import PrimitiveBase
from annadb.data_types.vector import VectorBase


class IDs(VectorBase):
    prefix = RESPONSE_IDS


class OkResponse(VectorBase):
    prefix = OK_RESPONSE


class GetMeta(MapUnique):
    prefix = GET_META

    @property
    def count(self):
        return self["count"]


class InsertMeta(MapUnique):
    prefix = INSERT_META

    @property
    def count(self):
        return self["count"]


class FindMeta(MapUnique):
    prefix = FIND_META

    @property
    def count(self):
        return self["count"]


class UpdateMeta(MapUnique):
    prefix = UPDATE_META

    @property
    def count(self):
        return self["count"]


class DeleteMeta(MapUnique):
    prefix = DELETE_META

    @property
    def count(self):
        return self["count"]


class QueryResponse(MapUnique):
    prefix = QUERY_RESPONSE

    @property
    def data(self):
        return self["data"]

    @property
    def meta(self):
        return self["meta"]


class Objects(MapUnique):
    prefix = RESPONSE_OBJECTS


class ResponsePrimitive(PrimitiveBase):
    prefix = RESPONSE_PRIMITIVE


class ErrorResponse(str, PrimitiveBase):
    prefix = ERROR_RESPONSE



class SimpleConsole:
    def __init__(self, title, console):
        self.title = title
        self.s = ""
        self.tabs = 0
        self.console = console

    def plus_tab(self):
        self.tabs += 1

    def minus_tab(self):
        self.tabs -= 1

    def add_text(self, text):
        self.s += text

    def new_line(self):
        self.s += "\n" + self.tabs * "\t"

    def print(self):
        # with self.console.pager():
        self.console.rule(self.title, style="cyan")
        self.console.print(self.s)
        self.console.print("")
        self.console.rule(style="blue")


class HTMLConsole:
    def __init__(self):
        self.s = ""
        self.tabs = 0

    def plus_tab(self):
        self.tabs += 1

    def minus_tab(self):
        self.tabs -= 1

    def add_text(self, text):
        self.s += text

    def new_line(self):
        self.s += "\n" + self.tabs * "\t"

    def out(self):
        self.s = "<pre><code>" + self.s + "</code></pre>"
        return self.s



from annadb.constants import QUERY_SET
from annadb.data_types.vector import VectorBase


class QuerySet(VectorBase):
    prefix = QUERY_SET

    def __init__(self, *args, collection: str = ""):
        self.collection = collection
        super(QuerySet, self).__init__(*args)

        self.queries = list(args)

    def push(self, query):
        self.queries.append(query)



from typing import List, Optional

from annadb.constants import STORAGE_VECTOR
from annadb.data_types.tyson import TysonItem


class VectorBase(TysonItem):
    prefix_color = "blue"
    instance_type = "vector"

    def __init__(self, *args, value: Optional[List[TysonItem]] = None):
        if args:
            self._value = list(args)
        elif value:
            self._value = value
        else:
            self._value = []

    def __getitem__(self, item):
        return self._value[item]

    def __eq__(self, other):
        if isinstance(other, list):
            return self._value == other
        if isinstance(other, VectorBase):
            return self._value == other._value and self.prefix == other.prefix
        return False

    def pretty(self, console):
        console.add_text(f"{self.pretty_prefix}[")
        console.plus_tab()
        for item in self._value:
            console.new_line()
            item.pretty(console)
            console.add_text(",")
        console.minus_tab()
        console.new_line()
        console.add_text("]")

    def to_html(self, console):
        console.add_text(f"{self.html_prefix}[")
        console.plus_tab()
        for item in self._value:
            console.new_line()
            item.to_html(console)
            console.add_text(",")
        console.minus_tab()
        console.new_line()
        console.add_text("]")


class Vector(VectorBase):
    prefix = STORAGE_VECTOR



from annadb.constants import SET_OPERATOR, INC_OPERATOR
from annadb.data_types.map import MapUnique


class Set(MapUnique):
    """
    Set({"key": "value"})
    """

    prefix = SET_OPERATOR


class Inc(MapUnique):
    prefix = INC_OPERATOR



from annadb.constants import (
    EQ_OPERATOR,
    NEQ_OPERATOR,
    GT_OPERATOR,
    GTE_OPERATOR,
    LT_OPERATOR,
    LTE_OPERATOR,
    AND_OPERATOR,
    OR_OPERATOR,
    NOT_OPERATOR,
)
from annadb.data_types.map import MapUnique
from annadb.data_types.modifier import ModifierBase
from annadb.data_types.vector import VectorBase


class BaseFindOperator:
    def __and__(self, other):
        return And(self, other)

    def __or__(self, other):
        return Or(self, other)


class Eq(BaseFindOperator, MapUnique):
    prefix = EQ_OPERATOR


class Neq(BaseFindOperator, MapUnique):
    prefix = NEQ_OPERATOR


class Gt(BaseFindOperator, MapUnique):
    prefix = GT_OPERATOR


class Gte(BaseFindOperator, MapUnique):
    prefix = GTE_OPERATOR


class Lt(BaseFindOperator, MapUnique):
    prefix = LT_OPERATOR


class Lte(BaseFindOperator, MapUnique):
    prefix = LTE_OPERATOR


class And(BaseFindOperator, VectorBase):
    prefix = AND_OPERATOR


class Or(BaseFindOperator, VectorBase):
    prefix = OR_OPERATOR


class Not(BaseFindOperator, ModifierBase):
    prefix = NOT_OPERATOR



import typing

if typing.TYPE_CHECKING:
    from annadb.connection import Connection
    from annadb.data_types.journal import QuerySet


class Transaction:
    def __init__(self, conn: "Connection"):
        self.conn = conn
        self.steps = []

    def push(self, step: "QuerySet"):
        self.steps.append(step.query_set)

    def run(self):
        return self.conn.send_query(*self.steps)



from abc import abstractmethod, ABC


class TysonItem(ABC):

    prefix = "undefined"
    prefix_color = "cyan"
    instance_type = ""

    @property
    def pretty_prefix(self):
        return f"[{self.prefix_color}]{self.prefix}[/{self.prefix_color}]"

    @property
    def html_prefix(self):
        return (
            f'<span class="prefix_{self.instance_type}">{self.prefix}</span>'
        )

    @abstractmethod
    def pretty(self, console):
        raise NotImplementedError



import typing

from annadb.data_types.journal import QuerySet
from annadb.query.base import FindInterface
from annadb.query.types import Sort

if typing.TYPE_CHECKING:
    from annadb.collection import Collection


class SortQuery(FindInterface):
    def __init__(self, *args, query_set: QuerySet, collection: "Collection"):
        self.collection = collection
        self.query_set = query_set
        self.query_set.push(Sort(*args))



from typing import Optional

from annadb.constants import PATH_TO_VALUE, ROOT
from annadb.data_types.primitive import PrimitiveBase
from annadb.query.find.operators import Eq, Gt, Gte, Lt, Lte, Neq
from annadb.query.sort.operators import Asc, Desc


class Path(PrimitiveBase):
    def __init__(self, value: Optional[str] = None):
        if value is None:
            self.prefix = ROOT
        else:
            self.prefix = PATH_TO_VALUE
        super(Path, self).__init__(value)

    def __getitem__(self, item):
        """
        Get sub field

        :param item: name of the subfield
        :return: Path
        """
        if self.value is not None:
            return Path(f"{self.value}.{item}")
        else:
            return Path(f"{item}")

    def __getattr__(self, item):
        """
        Get sub field

        :param item: name of the subfield
        :return: Path
        """
        if self.value is not None:
            return Path(f"{self.value}.{item}")
        else:
            return Path(f"{item}")

    def __eq__(self, other):
        return Eq({self: other})

    def __gt__(self, other):
        return Gt({self: other})

    def __ge__(self, other):
        return Gte({self: other})

    def __lt__(self, other):
        return Lt({self: other})

    def __le__(self, other):
        return Lte({self: other})

    def __ne__(self, other):
        return Neq({self: other})

    def __pos__(self):
        return Asc(self)

    def __neg__(self):
        return Desc(self)

    def __hash__(self):
        return hash("__root__")


root = Path()



from uuid import UUID

from annadb.constants import (
    STRING,
    NUMBER,
    BOOL,
    NULL,
    DELETED,
    COLLECTION_NAME,
    UTS_PREFIX,
)
from annadb.data_types.tyson import TysonItem


class PrimitiveBase(TysonItem):
    prefix_color = "yellow"
    value_color = "cyan"
    instance_type = "primitive"

    def __init__(self, value):
        self.value = value

    @property
    def pretty_value(self):
        return f"[{self.value_color}]{self.value}[/{self.value_color}]"

    @property
    def html_value(self):
        return f'<span class="value_{self.instance_type}">{self.value}</span>'

    def pretty(self, console):
        if self.value:
            console.add_text(f"{self.pretty_prefix}|{self.pretty_value}|")
        else:
            console.add_text(f"{self.pretty_prefix}")

    def to_html(self, console):
        if self.value:
            console.add_text(f"{self.html_prefix}|{self.html_value}|")
        else:
            console.add_text(f"{self.html_prefix}")

    def __hash__(self):
        return hash(self.prefix) + hash(self.value)

    def __eq__(self, other):
        if isinstance(other, PrimitiveBase):
            return self.prefix == other.prefix and self.value == other.value
        return False


class String(str, PrimitiveBase):
    value_color = "bright_green"
    prefix = STRING
    instance_type = "string"

    @classmethod
    def from_serialized(cls, value: str):
        value = value.replace("\\\\", "\\").replace("\\|", "|")
        return cls(value)


class Number(float, PrimitiveBase):
    value_color = "bright_blue"
    prefix = NUMBER
    instance_type = "number"


class UTS(int, PrimitiveBase):
    value_color = "bright_blue"
    prefix = UTS_PREFIX
    instance_type = "number"


class Bool(PrimitiveBase):
    value_color = "bright_red"
    prefix = BOOL
    instance_type = "bool"

    def __init__(self, value):
        if not isinstance(value, bool):
            if value == "true":
                value = True
            elif value == "false":
                value = False
            else:
                raise TypeError("Not bool")
        super(Bool, self).__init__(value)

    def __eq__(self, other):
        if isinstance(other, bool):
            return self.value is other
        return super(Bool, self).__eq__(other)


class Null(PrimitiveBase):
    prefix = NULL
    instance_type = "null"

    def __init__(self, _=None):
        super(Null, self).__init__(None)

    def __eq__(self, other):
        if other is None:
            return True
        if isinstance(other, PrimitiveBase):
            return self.prefix == other.prefix
        return False


class Deleted(PrimitiveBase):
    prefix = DELETED

    def __init__(self, _=None):
        super(Deleted, self).__init__(None)


class CollectionName(PrimitiveBase):
    prefix = COLLECTION_NAME


class Link(PrimitiveBase):
    prefix = ""
    instance_type = "link"

    def __init__(self, value, prefix):
        self.prefix = prefix
        value = UUID(value)
        super(Link, self).__init__(value)



from types import NoneType

from annadb.exceptions import ConvertError
from annadb.data_types.map import Map, MapBase, MapUnique, MapNotUnique
from annadb.data_types.modifier import ModifierBase
from annadb.data_types.primitive import (
    Number,
    String,
    Bool,
    Null,
    PrimitiveBase,
)
from annadb.data_types.tyson import TysonItem
from annadb.data_types.vector import Vector, VectorBase
from annadb.query.query_set import QuerySet

TYPE_TO_TYSON = {
    int: lambda o: Number(o),
    float: lambda o: Number(o),
    str: lambda o: String(o),
    bool: lambda o: Bool(o),
    NoneType: lambda o: Null(o),
}


def to_tyson(item):
    if isinstance(item, list):
        return Vector([to_tyson(i) for i in item])
    if isinstance(item, dict):
        return Map({to_tyson(k): to_tyson(v) for k, v in item.items()})
    if isinstance(item, tuple(TYPE_TO_TYSON.keys())):
        return TYPE_TO_TYSON[type(item)](item)
    if isinstance(item, TysonItem):
        if isinstance(item, VectorBase):
            item._value = [to_tyson(o) for o in item._value]
        if isinstance(item, MapBase):
            item.items = [
                (to_tyson(k), to_tyson(v)) for k, v in item.items.items()
            ]
        if isinstance(item, ModifierBase):
            item.value = to_tyson(item.value)
        return item
    raise ConvertError(f"Type {type(item)} is not TySON serializable")


def to_str(data):
    # Extra types
    if isinstance(data, QuerySet):
        res = []
        for query in data.queries:
            res.append(to_str(query))
        serialized_queries = "q[" + ",".join(res) + "]"
        return f"collection|{data.collection}|:{serialized_queries}"

    # TySON types
    if isinstance(data, String):
        data.value = data.value.replace("\\", "\\\\").replace("|", "\\|")
        return f"{data.prefix}|{data.value}|"

    if isinstance(data, Bool):
        if data.value is True:
            return f"{data.prefix}|true|"
        else:
            return f"{data.prefix}|false|"

    if isinstance(data, PrimitiveBase):
        if data.value:
            return f"{data.prefix}|{data.value}|"
        else:
            return f"{data.prefix}"

    if isinstance(data, ModifierBase):
        return f"{data.prefix}({to_str(data.value)})"

    if isinstance(data, VectorBase):
        res = f"{data.prefix}["
        for item in data._value:
            res += f"{to_str(item)},"
        res += "]"
        return res

    if isinstance(data, MapUnique):
        res = f"{data.prefix}{{"
        for k, v in data.items():
            res += f"{to_str(k)}:{to_str(v)},"
        res += "}"
        return res

    if isinstance(data, MapNotUnique):
        res = f"{data.prefix}{{"
        for item in data.items():
            res += f"{to_str(item.key)}:{to_str(item.value)},"
        res += "}"
        return res

    # Native types
    if data is None:
        return "null"
    if isinstance(data, bool):
        if data:
            return "b|true|"
        return "b|false|"
    if isinstance(data, int):
        return f"n|{data}|"
    if isinstance(data, float):
        return f"n|{data}|"
    if isinstance(data, str):
        data = data.replace("\\", "\\\\").replace("|", "\\|")
        return f"s|{data}|"
    if isinstance(data, list):
        res = "v["
        for item in data:
            res += f"{to_str(item)},"
        res += "]"
        return res
    if isinstance(data, dict):
        res = "m{"
        for k, v in data.items():
            res += f"{to_str(k)}:{to_str(v)},"
        res += "}"
        return res



from collections import OrderedDict
from typing import List, Optional

from annadb.constants import STORAGE_MAP
from annadb.data_types.pair import Pair
from annadb.data_types.tyson import TysonItem


class MapBase(TysonItem):
    prefix_color = "red"
    instance_type = "map"

    def __eq__(self, other):
        if isinstance(other, MapBase):
            return self.prefix == other.prefix and self._value == other._value
        return False

    def __len__(self):
        return len(self._value)


class MapUnique(MapBase):
    def __init__(self, data: Optional[dict] = None):
        if data is None:
            self._value = OrderedDict()
        else:
            self._value = data

    def __setitem__(self, key, value):
        self._value[key] = value

    def __getitem__(self, item):
        return self._value[item]

    def values(self):
        return list(self._value.values())

    def pretty(self, console):
        console.add_text(f"{self.pretty_prefix}{{")
        console.plus_tab()
        for k, v in self.items():
            console.new_line()
            k.pretty(console)
            console.add_text(":")
            v.pretty(console)
            console.add_text(",")
        console.minus_tab()
        console.new_line()
        console.add_text("}")

    def to_html(self, console):
        console.add_text(f"{self.html_prefix}{{")
        console.plus_tab()
        for k, v in self.items():
            console.new_line()
            k.to_html(console)
            console.add_text(":")
            v.to_html(console)
            console.add_text(",")
        console.minus_tab()
        console.new_line()
        console.add_text("}")

    def items(self):
        return self._value.items()


class MapNotUnique(MapBase):
    def __init__(self, data: Optional[List[Pair]] = None):
        if data is None:
            self._value: List[Pair] = []
        else:
            self._value = data

    def __setitem__(self, key, value):
        self._value.append(Pair(key, value))

    def __getitem__(self, item):
        return self._value[item]

    def pretty(self, console):
        console.add_text(f"{self.pretty_prefix}{{")
        console.plus_tab()
        for item in self._value:
            console.new_line()
            item.key.pretty(console)
            console.add_text(":")
            item.value.pretty(console)
            console.add_text(",")
        console.minus_tab()
        console.new_line()
        console.add_text("}")

    def items(self):
        return self._value


class Map(MapUnique):
    prefix = STORAGE_MAP

    def __eq__(self, other):
        if isinstance(other, dict):
            return self._value == other
        return super(Map, self).__eq__(other)



import typing

from annadb.query.base import BaseQuery
from annadb.query.query_set import QuerySet
from annadb.query.types import Delete

if typing.TYPE_CHECKING:
    from annadb.collection import Collection


class DeleteQuery(BaseQuery):
    def __init__(
        self,
        collection: "Collection",
        query_set: typing.Optional[QuerySet] = None,
    ):
        self.collection = collection
        query = Delete()
        if query_set is None:
            self.query_set = QuerySet(query, collection=self.collection.name)
        else:
            query_set.push(query)
            self.query_set = query_set



from annadb.constants import (
    FIND_QUERY,
    GET_QUERY,
    INSERT_QUERY,
    SORT_QUERY,
    UPDATE_QUERY,
    OFFSET_QUERY,
    LIMIT_QUERY,
    DELETE_QUERY,
)
from annadb.data_types.modifier import ModifierBase
from annadb.data_types.primitive import PrimitiveBase
from annadb.data_types.vector import VectorBase


class Find(VectorBase):
    prefix = FIND_QUERY


class Get(VectorBase):
    prefix = GET_QUERY


class Insert(VectorBase):
    prefix = INSERT_QUERY


class Sort(VectorBase):
    prefix = SORT_QUERY


class Update(VectorBase):
    prefix = UPDATE_QUERY


class Offset(ModifierBase):
    prefix = OFFSET_QUERY


class Limit(ModifierBase):
    prefix = LIMIT_QUERY


class Delete(PrimitiveBase):
    prefix = DELETE_QUERY

    def __init__(self, _=None):
        super(Delete, self).__init__(None)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['annadb',
 'annadb.data_types',
 'annadb.query',
 'annadb.query.delete',
 'annadb.query.find',
 'annadb.query.get',
 'annadb.query.insert',
 'annadb.query.limit',
 'annadb.query.offset',
 'annadb.query.sort',
 'annadb.query.update',
 'annadb.shell']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'pyzmq>=23.1.0,<24.0.0',
 'textual>=0.1.18,<0.2.0',
 'tyson>=0.1.5,<0.2.0']

entry_points = \
{'console_scripts': ['annadb = annadb.shell:shell']}

setup_kwargs = {
    'name': 'annadb',
    'version': '0.0.2',
    'description': 'AnnaDB driver and shell',
    'long_description': None,
    'author': 'Roman Right',
    'author_email': 'roman-right@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import typing

from annadb.query.base import BaseQuery
from annadb.query.query_set import QuerySet
from annadb.query.types import Insert
from annadb.response import ErrorResponse

if typing.TYPE_CHECKING:
    from annadb.collection import Collection


class InsertQuery(BaseQuery):
    def __init__(self, *args, collection: "Collection"):
        self.collection = collection
        self.query_set = QuerySet(
            Insert(*args), collection=self.collection.name
        )


class InsertOneQuery(InsertQuery):
    def __init__(self, item, collection: "Collection"):
        super(InsertOneQuery, self).__init__(item, collection=collection)

    def run(self):
        res = super(InsertOneQuery, self).run()
        if not isinstance(res, ErrorResponse):
            res["data"] = res.data[0]
        return res



from tyson import deserialize

from annadb.console import SimpleConsole, HTMLConsole
from annadb.data_types.pair import Pair
from annadb.data_types.types_map import build_primitive, factory


class Journal:
    def __init__(self, data):
        self._value = data

    def __getitem__(self, item):
        return self._value[item]

    @classmethod
    def deserialize(cls, query: str):
        res = deserialize(query)
        return cls(
            [Pair(build_primitive(o[0]), factory(o[1])) for o in res.data]
        )

    def pretty(self, topic, console):
        console = SimpleConsole(topic, console)
        for pair in self._value:
            console.new_line()
            pair.key.pretty(console)
            console.add_text(":")
            pair.value.pretty(console)
            console.add_text(";")
        console.print()

    def to_html(self):
        console = HTMLConsole()
        for pair in self._value:
            pair.key.to_html(console)
            console.add_text(":")
            pair.value.to_html(console)
            console.add_text(";")
            console.new_line()
        return console.out()



import typing

from annadb.query.delete.query import DeleteQuery
from annadb.query.find.query import FindQuery
from annadb.query.get.query import GetQuery, GetOneQuery
from annadb.query.insert.query import InsertQuery, InsertOneQuery

if typing.TYPE_CHECKING:
    from annadb.connection import Connection


class Collection:
    def __init__(self, name: str, conn: "Connection"):
        self.name = name
        self.conn = conn

    def get(self, *args):
        return GetQuery(*args, collection=self)

    def get_one(self, id):
        return GetOneQuery(id, collection=self)

    def find(self, *args):
        return FindQuery(*args, collection=self)

    def all(self):
        return FindQuery(collection=self)

    def insert(self, *args):
        return InsertQuery(*args, collection=self)

    def insert_one(self, item):
        return InsertOneQuery(item, collection=self)

    def delete(self):
        return DeleteQuery(collection=self)



STRING: str = "s"
NUMBER: str = "n"
UTS_PREFIX: str = "uts"
BOOL: str = "b"
NULL: str = "null"
COLLECTION_NAME: str = "collection"
PATH_TO_VALUE: str = "value"
DELETED: str = "deleted"

STORAGE_MAP: str = "m"

STORAGE_VECTOR: str = "v"

# QUERIES
QUERY_SET: str = "q"
INSERT_QUERY: str = "insert"
FIND_QUERY: str = "find"
GET_QUERY: str = "get"
UPDATE_QUERY: str = "update"
DELETE_QUERY: str = "delete"
SORT_QUERY: str = "sort"
OFFSET_QUERY: str = "offset"
LIMIT_QUERY: str = "limit"

# FIND OPERATORS
EQ_OPERATOR: str = "eq"
NEQ_OPERATOR: str = "neq"
GT_OPERATOR: str = "gt"
GTE_OPERATOR: str = "gte"
LT_OPERATOR: str = "lt"
LTE_OPERATOR: str = "lte"

AND_OPERATOR: str = "and"
OR_OPERATOR: str = "or"

NOT_OPERATOR: str = "not"

# UPDATE OPERATORS
SET_OPERATOR: str = "set"
INC_OPERATOR: str = "inc"

# SORT OPERATORS
ASC_OPERATOR: str = "asc"
DESC_OPERATOR: str = "desc"

# RESPONSE
RESPONSE_OBJECTS: str = "objects"
RESPONSE_IDS: str = "ids"
RESPONSE: str = "data"
RESPONSE_PRIMITIVE: str = "result"
OK_RESPONSE: str = "ok"
ERROR_RESPONSE: str = "error"
QUERY_RESPONSE: str = "response"
INSERT_META: str = "insert_meta"
GET_META: str = "get_meta"
UPDATE_META: str = "update_meta"
FIND_META: str = "find_meta"
DELETE_META: str = "delete_meta"

# OTHER
ROOT: str = "root"
INTERNAL_COLLECTION_NAME: str = "_internal"



import click
from rich.console import Console

from annadb.connection import Connection
from annadb.data_types.journal import Journal


@click.command()
@click.option(
    "--uri",
    required=True,
    type=str,
    help="Connection string",
)
def shell(uri):
    conn = Connection.from_connection_string(uri)
    console = Console(tab_size=4)
    console.print(
        r"""
      ___           ___           ___           ___           ___           ___
     /\  \         /\__\         /\__\         /\  \         /\  \         /\  \
    /::\  \       /::|  |       /::|  |       /::\  \       /::\  \       /::\  \
   /:/\:\  \     /:|:|  |      /:|:|  |      /:/\:\  \     /:/\:\  \     /:/\:\  \
  /::\~\:\  \   /:/|:|  |__   /:/|:|  |__   /::\~\:\  \   /:/  \:\__\   /::\~\:\__\
 /:/\:\ \:\__\ /:/ |:| /\__\ /:/ |:| /\__\ /:/\:\ \:\__\ /:/__/ \:|__| /:/\:\ \:|__|
 \/__\:\/:/  / \/__|:|/:/  / \/__|:|/:/  / \/__\:\/:/  / \:\  \ /:/  / \:\~\:\/:/  /
      \::/  /      |:/:/  /      |:/:/  /       \::/  /   \:\  /:/  /   \:\ \::/  /
      /:/  /       |::/  /       |::/  /        /:/  /     \:\/:/  /     \:\/:/  /
     /:/  /        /:/  /        /:/  /        /:/  /       \::/__/       \::/__/
     \/__/         \/__/         \/__/         \/__/         ~~            ~~
    """
    )

    while True:

        line = input(">>> ")
        if line:
            query = f"{line}\n"
            while line:
                line = console.input()
                query += f"{line}\n"

            if query == "":
                break

            try:
                request_doc = Journal.deserialize(query)
            except ValueError as e:
                console.print(str(e))
                continue

            console.clear()

            request_doc.pretty("Request", console)

            response_doc = conn.send_query(query, value_only=False)

            response_doc.pretty("Response", console)
        else:
            break

    conn.close()


if __name__ == "__main__":
    shell()



import typing

from annadb.data_types.journal import QuerySet
from annadb.query.base import BaseQuery
from annadb.query.types import Update

if typing.TYPE_CHECKING:
    from annadb.collection import Collection


class UpdateQuery(BaseQuery):
    def __init__(self, *args, query_set: QuerySet, collection: "Collection"):
        self.collection = collection
        query_set.push(Update(*args))
        self.query_set = query_set



from annadb.data_types.tyson import TysonItem


class ModifierBase(TysonItem):
    prefix_color = "green"
    instance_type = "modifier"

    def __init__(self, value):
        self.value = value

    def pretty(self, console):
        console.add_text(f"{self.pretty_prefix}(")
        self.value.pretty(console)
        console.add_text(")")

    def to_html(self, console):
        console.add_text(f"{self.html_prefix}(")
        self.value.to_html(console)
        console.add_text(")")
